#include "core/VulkanWrapper.h"

#include <core/Logger.h>
#include <core/Utility.h>

#include "GLFW.h"
#include "VulkanUtils.h"

// VkCommandPool commandPool;
// VkCommandBuffer* commandsBuffers;
// VkFramebuffer* framebuffers;
// VkSemaphore semaphore;
// VkSemaphore renderSemaphore;
// VkRenderPass renderPass;

static VkPhysicalDevice physicalDevice;
static u32 graphicsFamily = 0;
static u32 presentFamily = 0;
static VkDevice device;
static VkQueue graphicsQueue;
static VkQueue presentQueue;
static VkSurfaceFormatKHR surfaceFormat;
static VkSurfaceKHR surface;
static VkSwapchainKHR swapchain;
static VulkanSwapchainImages images;

static int getSurfaceFormatPoints(const VkSurfaceFormatKHR* sf) {
    if(sf->format == VK_FORMAT_B8G8R8A8_UNORM &&
       sf->colorSpace == VK_COLOR_SPACE_SRGB_NONLINEAR_KHR) {
        return 10;
    }
    return 1;
}

static int getSurfacePresentModePoints(VkPresentModeKHR m) {
    if(m == VK_PRESENT_MODE_MAILBOX_KHR) {
        return 5;
    } else if(m == VK_PRESENT_MODE_FIFO_KHR) {
        return 2;
    }
    return 0;
}

static bool getSwapchainSize(VkExtent2D* size) {
    VkSurfaceCapabilitiesKHR c = {0};
    VK_ASSERT(
        vkGetPhysicalDeviceSurfaceCapabilitiesKHR(physicalDevice, surface, &c));
    if(c.currentExtent.width != 0xFFFF'FFFFu &&
       c.currentExtent.height != 0xFFFF'FFFFu) {
        *size = c.currentExtent;
        LOG_INFO("Swapchain size: %ux%u", size->width, size->height);
        return false;
    }
    int w = 0;
    int h = 0;
    glfwGetFramebufferSize(getWindow(), &w, &h);
    if(w <= 0 || h <= 0) {
        LOG_ERROR("Could not get framebuffer size");
        return true;
    }
    LOG_INFO("Framebuffer size: %dx%d", w, h);
    size->width =
        clampU32((u32)w, c.minImageExtent.width, c.maxImageExtent.width);
    size->height =
        clampU32((u32)h, c.minImageExtent.height, c.maxImageExtent.height);
    LOG_INFO("Swapchain size: %ux%u", size->width, size->height);
    return false;
}

static bool initSwapchain() {
    VulkanSwapchainData d = {
        .physicalDevice = physicalDevice, .device = device, .surface = surface};
    if(getSwapchainSize(&d.size)) {
        LOG_ERROR("Could not retrieve any swapchain size");
        return true;
    }
    if(findVulkanSurfaceFormat(
           &d.surfaceFormat, physicalDevice, surface, getSurfaceFormatPoints)) {
        LOG_ERROR("Could not find surface format");
        return true;
    }
    surfaceFormat = d.surfaceFormat;
    if(findVulkanSurfacePresentMode(
           &d.presentMode, physicalDevice, surface,
           getSurfacePresentModePoints)) {
        LOG_ERROR("Could not find present mode");
        return true;
    }
    u32 queueFamilyIndices[] = {graphicsFamily, presentFamily};
    if(graphicsFamily != presentFamily) {
        d.sharingMode = VK_SHARING_MODE_CONCURRENT;
        d.queueFamilyIndexCount = 2;
        d.queueFamilyIndices = queueFamilyIndices;
    } else {
        d.sharingMode = VK_SHARING_MODE_EXCLUSIVE;
    }
    return initVulkanSwapchain(&swapchain, &d);
}

static bool initDevice() {
    graphicsFamily =
        findVulkanQueueFamily(physicalDevice, VK_QUEUE_GRAPHICS_BIT);
    presentFamily = findVulkanSurfaceQueueFamily(physicalDevice, surface);
    if(graphicsFamily == INVALID_VULKAN_QUEUE_FAMILY ||
       presentFamily == INVALID_VULKAN_QUEUE_FAMILY) {
        return true;
    }
    bool same = graphicsFamily == presentFamily;
    VulkanDeviceQueueData data[] = {
        {graphicsFamily, 1.0f}, {presentFamily, 1.0f}};
    const char* extensions[] = {VK_KHR_SWAPCHAIN_EXTENSION_NAME};
    if(initVulkanDevice(
           &device, physicalDevice, data, same ? 1 : 2, extensions,
           ARRAY_LENGTH(extensions))) {
        return true;
    }
    vkGetDeviceQueue(device, graphicsFamily, 0, &graphicsQueue);
    if(graphicsQueue == VK_NULL_HANDLE) {
        LOG_ERROR("Cannot get device graphics queue");
        return true;
    }
    if(same) {
        presentQueue = graphicsQueue;
        return false;
    }
    vkGetDeviceQueue(device, presentFamily, 0, &presentQueue);
    if(presentQueue == VK_NULL_HANDLE) {
        LOG_ERROR("Cannot get device present queue");
        return true;
    }
    return false;
}

static int getDevicePoints(VkPhysicalDevice pd) {
    int points = 0;
    VkPhysicalDeviceProperties p;
    vkGetPhysicalDeviceProperties(pd, &p);
    LOG_INFO("Checking '%s'", p.deviceName);
    switch(p.deviceType) {
        case VK_PHYSICAL_DEVICE_TYPE_DISCRETE_GPU: points += 100; break;
        case VK_PHYSICAL_DEVICE_TYPE_INTEGRATED_GPU: points += 50; break;
        case VK_PHYSICAL_DEVICE_TYPE_VIRTUAL_GPU: points += 20; break;
        default: break;
    }
    u32 gf = findVulkanQueueFamily(pd, VK_QUEUE_GRAPHICS_BIT);
    if(gf == INVALID_VULKAN_QUEUE_FAMILY) {
        LOG_INFO("> ... has no graphics family");
        points = -1;
    }
    u32 pf = findVulkanSurfaceQueueFamily(pd, surface);
    if(pf == INVALID_VULKAN_QUEUE_FAMILY) {
        LOG_INFO("> ... has no present family");
        points = -1;
    }
    if(!hasVulkanExtension(pd, VK_KHR_SWAPCHAIN_EXTENSION_NAME)) {
        LOG_INFO("> ... has no swapchain support");
        points = -1;
    }
    VkSurfaceFormatKHR sf = {0};
    if(findVulkanSurfaceFormat(&sf, pd, surface, getSurfaceFormatPoints)) {
        LOG_INFO("> ... has no matching surface format");
        points = -1;
    } else {
        points += getSurfaceFormatPoints(&sf);
    }
    VkPresentModeKHR m = 0;
    if(findVulkanSurfacePresentMode(
           &m, pd, surface, getSurfacePresentModePoints)) {
        LOG_INFO("> ... has no matching present mode");
        points = -1;
    } else {
        points += getSurfacePresentModePoints(m);
    }
    LOG_INFO("> Final points: %d", points);
    return points;
}

static bool initPhysicalDevice() {
    LOG_INFO("Searching for physical devices ...");
    if(findVulkanPhysicalDevice(&physicalDevice, getDevicePoints)) {
        LOG_ERROR("No matching physical device was found");
        return true;
    }
    VkPhysicalDeviceProperties p;
    vkGetPhysicalDeviceProperties(physicalDevice, &p);
    LOG_INFO("Best Device: %s", p.deviceName);
    return false;
}

static bool initSwapchainImages() {
    if(initVulkanSwapchainImages(
           &images, device, swapchain, surfaceFormat.format)) {
        LOG_ERROR("Could not get swapchain images");
        return true;
    }
    LOG_INFO("Found %u images", images.amount);
    return false;
}

bool initVulkan() {
    return initVulkanInstance() || initVulkanDebugging() ||
           initVulkanSurface(&surface, getWindow()) || initPhysicalDevice() ||
           initDevice() || initSwapchain() || initSwapchainImages();
    /*  || initSemaphore(&vk.semaphore) ||
       initSemaphore(&vk.renderSemaphore) || initCommandBuffers() ||
       initRenderPass() || initFramebuffers()*/
}

void renderVulkan() {
}

void destroyVulkan(void) {
    // if(vk.device.logical != VK_NULL_HANDLE) {
    //     destroyFramebuffers();
    //     vkDestroyRenderPass(vk.device.logical, vk.renderPass, nullptr);
    //     destroyCommandBuffers();
    //     destroySemaphore(vk.semaphore);
    //     destroySemaphore(vk.renderSemaphore);
    // }
    destroyVulkanSwapchainImages(&images, device);
    destroyVulkanSwapchain(swapchain, device);
    destroyVulkanDevice(device);
    destroyVulkanSurface(surface);
    destroyVulkanDebugging();
    destroyVulkanInstance();
}
